/*---------------------------------------------------------------------------*\
 =========                 |
 \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
 \\    /   O peration     |
 \\  /    A nd           | Copyright (C) 2016 OpenFOAM Foundation
 \\/     M anipulation  |
 -------------------------------------------------------------------------------
 License
 This file is part of OpenFOAM.
 
 OpenFOAM is free software: you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 for more details.
 
 You should have received a copy of the GNU General Public License
 along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.
 
 \*---------------------------------------------------------------------------*/

#include "moodyCore.H"
#include "addToRunTimeSelectionTable.H"
#include "sixDoFRigidBodyMotion.H"
#include "moodyWrapper.h"

// #include "Time.H"

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

namespace Foam
{
    namespace sixDoFRigidBodyMotionRestraints
    {
        defineTypeNameAndDebug(moodyCore, 0);
        
        addToRunTimeSelectionTable
        (
         sixDoFRigidBodyMotionRestraint,
         moodyCore,
         dictionary
         );
    }
}



// * * * Private function * * * //
bool Foam::sixDoFRigidBodyMotionRestraints::moodyCore::updateNumberOfPoints()
{
    int newPointSize = moodyGetNumberOfPoints();
    if ( nPoints_ == newPointSize )
        return true;
    
    // Else update container sizes:
    nPoints_=newPointSize;
    mooringPositions_.resize(nPoints_*3);
    UFluid_.resize(nPoints_*3);
    dUdtFluid_.resize(nPoints_*3);
    rhoFluid_.resize(nPoints_);
    
    return false;
}


//- Compute a linear mooringRamp based on current time value.
Foam::scalar Foam::sixDoFRigidBodyMotionRestraints::moodyCore::mooringRamp(scalar t) const
{
    scalar moorRamp = 1.0;
    
    if ( mooringRampTime_>0.0 )
    {
        moorRamp = max(
                       0.0,
                       min(
                           (t-startTime_)/(mooringRampTime_),
                           1.0
                           )
                       );
    }
    if (moorRamp < 1.0)
        Info << "MoodyCore::moorin ramp is active, factor: " << moorRamp << endl;
    
    return moorRamp;
}

// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

Foam::sixDoFRigidBodyMotionRestraints::moodyCore::moodyCore
(
 const word& name,
 const dictionary& sDoFRBMRDict
 )
:
sixDoFRigidBodyMotionRestraint(name, sDoFRBMRDict),
attachPts_(),
mooringSystemFilename_(),
noOfAttachments_(),
nPoints_(0),
mooringPositions_(),
UFluid_(),
dUdtFluid_(),
rhoFluid_(),
mooringRampTime_(),
startTime_(0.0),
moored_(false),
moodyCoreStarted_(false)
{
    read(sDoFRBMRDict);
}

// * * * * * * * * * * * * * * * * Destructor  * * * * * * * * * * * * * * * //

Foam::sixDoFRigidBodyMotionRestraints::moodyCore::~moodyCore()
{
    if (moodyCoreStarted_)
        moodyClose();
}


// * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * * //
void Foam::sixDoFRigidBodyMotionRestraints::moodyCore::restrain
(
 const sixDoFRigidBodyMotion& motion,
 vector& restraintPosition,
 vector& restraintForce,
 vector& restraintMoment
 )
{
    // Do the first time: Open up a moodyCore API object //
    if ( !moodyCoreStarted_ && moored_)
    {
        std::vector<std::string> inputFlags;
        // Add the input file name to the flags.
	// Output name specified in moodyCore input file (
	// API.output = resultName)
        inputFlags.push_back("-f");
        inputFlags.push_back(mooringSystemFilename_);
        
        // Generate and fill the formated inputs std::vector<double> attachInfo for moody.h
        std::vector<double> attachInfo(3*noOfAttachments_); // Always 3D!
        point startAttachPoint;
        
        Info << "Initialising mooring software MoodyCore" << endl;
        Info << "MoodyCore: noOfAttachments "  << noOfAttachments_ << endl;
        
        for ( int ii = 0; ii<noOfAttachments_; ++ii)
        {
            // Transform the starting points to last motion state
            startAttachPoint = motion.transform(attachPts_[ii]);
            Info << "MoodyCore: Start attach point" << ii <<": " << startAttachPoint << endl;
            // Fill attachInfo with values //
            attachInfo[ii*3] = startAttachPoint[0];
            attachInfo[ii*3+1] = startAttachPoint[1];
            attachInfo[ii*3+2] = startAttachPoint[2];
        }
        // Call MoodyCore
        moodyInit(mooringSystemFilename_.c_str(), 3*noOfAttachments_ , &(attachInfo[0]) , motion.time().value());
        
        moodyCoreStarted_ = true;
        
        // End with setting the size of all fluid containers.
        updateNumberOfPoints();
        
        return;
    }
    else if (moored_)
    {
        // do for each new time:
        double t = motion.time().value();
        double tPrev = t - motion.time().deltaTValue();
        
        Info << "Restraining body" << endl;
        // Compute values at present motion state
        
        List<point> currentPositions(attachPts_);
        
        std::vector<double> attachInfo(3*noOfAttachments_,0.0); // Always 3D!
        std::vector<double> allForces(3*noOfAttachments_,0.0); // Always 3D!
	
	for ( int ii = 0; ii<noOfAttachments_; ++ii)
        {
            // Transform the starting points to present motion state
            currentPositions[ii] = motion.transform(attachPts_[ii]);
            
            // Fill attachInfo with transformed values
            attachInfo[ii*3] =   currentPositions[ii].x();
            attachInfo[ii*3+1] = currentPositions[ii].y();
            attachInfo[ii*3+2] = currentPositions[ii].z();
        }
        
        // Send info to moody
        Info << "Solving mooring system between times : " << tPrev  << " and " << t <<endl;
        moodySolve(&(attachInfo[0]) , &(allForces[0]) ,tPrev , t);
        Info << "Mooring forces computed!" << endl;
        
        //- Compute the mooring ramping factor \in [0,1]
        scalar moorRamp = mooringRamp(t);
        
        // Extract info from allForces and compute moment and total force
        vector force(vector::zero);
        vector moment(vector::zero);
        
        
        // Set force and moment for each attachment point.
        for ( int ii = 0; ii<noOfAttachments_; ++ii)
        {
            // Negative sign to account for outward pointing normal direction of forces
            force[0] = -moorRamp*allForces[ii*3];
            force[1] = -moorRamp*allForces[ii*3+1];
            force[2] = -moorRamp*allForces[ii*3+2];
            
            // NB: The force and the moments of each force
            moment= (currentPositions[ii]-motion.centreOfRotation())^force;
            
	    restraintForce += force;
            restraintMoment += moment; 
            
	    // Accumulate the force for the restrained body
            Info << "attachment point " << ii << ", fairlead at " << currentPositions[ii] <<
            ": (force, N) " << force << " (moment, Nm) " << moment << endl;
        }
        
        // NB: End with setting the number of points in the latest configuration.
	//     Only needed for use with MoodyCore's mesh adaptivity.
        updateNumberOfPoints();
        
    }
    // Finally: set mooring point position to centerOfRotation to avoid extra moment computation in appyRestraints()
    restraintPosition = motion.centreOfRotation();
    
}


int Foam::sixDoFRigidBodyMotionRestraints::moodyCore::getRestraintPoints(Foam::List<Foam::point>& restraintPoints)
{
    if (!moodyCoreStarted_)
        return 0;
    
    //    nPoints_ = moodyGetNumberOfPoints(); // update number of points
    //    std::vector<double> moodyPoints(nPoints_*3);
    // fill position vector
    double* mp = &mooringPositions_[0]; //dyPoints[0];
    moodyGetPositions(nPoints_,mp);
    
    // List<point> restraintPoints(nPoints_);
    restraintPoints.setSize(nPoints_);
    
    // fill point list
    forAll(restraintPoints, i)
    {
        restraintPoints[i].x() = *mp++;
        restraintPoints[i].y() = *mp++;
        restraintPoints[i].z() = *mp++;
    }
    return nPoints_;
}

void Foam::sixDoFRigidBodyMotionRestraints::moodyCore::setFluidState(scalar t,const List<vector>& U , const List<vector>& dUdt, const List<scalar>& rho)
{
    if (!moodyCoreStarted_)
        return;
    
    // Create vector containers.
    // std::vector<double> vel(nPoints_*3), acc(nPoints_*3),density(nPoints_);
    double* vPtr=&(UFluid_[0]);
    double* aPtr=&(dUdtFluid_[0]);
    double* rhoPtr = &(rhoFluid_[0]);
    
    //- Copy the info to std::vector format to ensure store order. May work anyway? -//
    forAll(U,i)
    {
        *vPtr++ = U[i].x();
        *vPtr++ = U[i].y();
        *vPtr++ = U[i].z();
        
        *aPtr++ = dUdt[i].x();
        *aPtr++ = dUdt[i].y();
        *aPtr++ = dUdt[i].z();
        
        *rhoPtr++ = rho[i];
    }
    
    // Set fluid state in moody:
    moodySetFlow(t, nPoints_, &(UFluid_[0]),&(dUdtFluid_[0]),&(rhoFluid_[0]));
}


bool Foam::sixDoFRigidBodyMotionRestraints::moodyCore::read
(
 const dictionary& sDoFRBMRDict
 )
{
    sixDoFRigidBodyMotionRestraint::read(sDoFRBMRDict);
    
    attachPts_=sDoFRBMRCoeffs_.lookupOrDefault< List<point> >("attachmentPoints", List<point>());
    noOfAttachments_ = int(attachPts_.size()); 
    
    if( noOfAttachments_ > 0 )
    {
        sDoFRBMRCoeffs_.lookup("mooringSystemFilename") >> mooringSystemFilename_;
        
        mooringRampTime_=sDoFRBMRCoeffs_.lookupOrDefault<scalar>("mooringRampTime",0.0);
        mooringRampTime_=sDoFRBMRCoeffs_.lookupOrDefault<scalar>("rampStartTime",0.0);
        
        moored_ = true;
    }
    return true;
}


void Foam::sixDoFRigidBodyMotionRestraints::moodyCore::write
(
 Ostream& os
 ) const
{
    
    sixDoFRigidBodyMotionRestraint::write(os);
    
    os.writeEntry("attachmentPoints",attachPts_);
    os.writeEntry("mooringSystemFilename",mooringSystemFilename_);
    os.writeEntry("mooringRampTime",mooringRampTime_);
    os.writeEntry("startTime",startTime_);
    
}


// ************************************************************************* //
