/*---------------------------------------------------------------------------*\
 =========                 |
 \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
 \\    /   O peration     |
 \\  /    A nd           | Copyright (C) 2016 OpenFOAM Foundation
 \\/     M anipulation  |
 -------------------------------------------------------------------------------
 License
 This file is part of OpenFOAM.
 
 OpenFOAM is free software: you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 for more details.
 
 You should have received a copy of the GNU General Public License
 along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.
 
 Class
 Foam::sixDoFRigidBodyMotionRestraints::moodyCore
 
 Description
 sixDoFRigidBodyMotionRestraints moodyCore.
 Applies dynamic mooring loads computed by moodyCore to a body
 
 SourceFiles
 moodyCore.C
 
 \*---------------------------------------------------------------------------*/

#ifndef moodyCore_H
#define moodyCore_H

#include "sixDoFRigidBodyMotionRestraint.H"
#include <vector>

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
    
    namespace sixDoFRigidBodyMotionRestraints
    {
        
        /*---------------------------------------------------------------------------*\
         Class moodyCore Declaration
         \*---------------------------------------------------------------------------*/
        
        class moodyCore
        :
        public sixDoFRigidBodyMotionRestraint
        {
	
	// Private data
        
        // Reference points of attachment to the solid body //
        List<point> attachPts_;
        
        // Word containing the filename of the input file for the mooring system //
        string mooringSystemFilename_;
        
        // The number of attachment points //
        int noOfAttachments_;
        
        // Number of fluid points sampled
        int nPoints_;
        
        // Containers for fluid state and mooring poisitions
        std::vector<double> mooringPositions_;
        std::vector<double> UFluid_;
        std::vector<double> dUdtFluid_;
        std::vector<double> rhoFluid_;
        
        // The ramp time of the mooring force
	//(not invoked in moody, but before force is returned to motion_ 
        scalar mooringRampTime_;
        
        // Last call time of function //
        scalar startTime_;
        
        // If moored or not
        bool moored_;
        
        // Switch to ensure that moody is only intialised once
        bool moodyCoreStarted_;
        
        // Private member functions
        
        /** Compute and return mooring ramp value at time t */
        scalar mooringRamp(scalar t) const;
        
        /** Update the number of points in the fluid state containers based on Moody information. */
        bool updateNumberOfPoints();
        
    public:
        
        //- Runtime type information
        TypeName("moodyCore");
        
        
        // Constructors
        
        //- Construct from components
        moodyCore
        (
         const word& name,
         const dictionary& sDoFRBMRDict
         );
        
        //- Construct and return a clone
        virtual autoPtr<sixDoFRigidBodyMotionRestraint> clone() const
        {
            return autoPtr<sixDoFRigidBodyMotionRestraint>
            (
             new moodyCore(*this)
             );
        }
        
        
        //- Destructor
        virtual ~moodyCore();
        
        
        // Member Functions
        virtual void restrain
        (
         const sixDoFRigidBodyMotion& motion,
         vector& restraintPosition,
         vector& restraintForce,
         vector& restraintMoment
         );
        
        // Get mooring  positions from moody. Called from meshMotion solver. //
        virtual int  getRestraintPoints(List<point>& allRestraintPoints);
        
        // Sends fluid state properties at quad points (from getRestraintPoints()) to moody for use in added mass and drag force.
        virtual void setFluidState(scalar t, const List<vector>& U ,const  List<vector>& Udot, const List<scalar>& rho);
        
        //- Update properties from given dictionary
        virtual bool read(const dictionary& sDoFRBMRDict);
        
        //- Write
        virtual void write(Ostream&) const;
    };
    
    
    // * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //
    
} // End namespace sixDoFRigidBodyMotionRestraints
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
