
    /* This file collects interpolated values of U,  dUdt and rho from all mooring points of Moody.
     * It relies on two new interface functions in rigidBodyModel (rigidBodyDynamics library):
     *     1. List<int> rigidBodyModel::getRestraintPoints(List<List<point>> allPoints)
     *         which collects a point list for each restraint. The number of points for each restraint is returned.
     *         The base class restraint returns 0 size for non-compatible restraints (i.e. all but moodyR)
     *     2. void rigidBodyModel::setFlow(const List<List<vector>>& U, const List<List<vector>>& dUdt, const List<List<scalar>>& rho)
     *         which sends the probed values to all restraints that so require. At present: only to moodyR.
     *
     * As MoodyCore only lives on the master node, both function 1 and 2 above are only working for Pstream::master().
     * The complete probed data is therefore only gathered to the master list.
     *
     * */
    
    
    // Get the U and rho from db():
    const volVectorField& U = db().lookupObject<volVectorField>("U"); 
    const volScalarField& rho = db().lookupObject<volScalarField>("rho"); 
    scalar rhoMax = max(rho.primitiveField());
    
    // Check if ddt-field exists.
    bool is_ddtU = db().found("ddtCorrDdt0(U)");
    
    // Create reference to fluid acceleration
    const volVectorField& ddtU = is_ddtU ? db().lookupObject<volVectorField>("ddtCorrDdt0(U)") : volVectorField(0*U);
    
    // Generate interpolation objects:
    interpolationCellPoint<vector> UInterp(U);
    interpolationCellPoint<scalar> rhoInterp(rho);
    interpolationCellPoint<vector> ddtUInterp(ddtU);
    
    forAll(motion_.allRestraints(),rI)
    {
        List<point> allPoints;
        int nPoints=0;
        if ( Pstream::master() )
	  {
            nPoints = motion_.allRestraints()[rI].getRestraintPoints(allPoints);
	  }
        Pstream::scatter(nPoints);
        Pstream::scatter(allPoints);
        
        if (nPoints > 0)
        {
	  // Probed data initialized with default values if out-of-cfd-domain
	  List<vector> UProbed(nPoints,vector::zero);
	  List<vector> dUdtProbed(nPoints,vector::zero);
	  List<scalar> rhoProbed(nPoints,rhoMax);//pTraits<scalar>::min);
            
	  // Cell list for parallell work
	  List<label> cellNos(nPoints);
            
            // find cell for all points in each restraint
            forAll(allPoints,ii)
            {
                const point pI = allPoints[ii];
                // find cell of this point
		label seedCelli = -1;  // -1 = no seed: should try with starting from the one found last time...
		bool useTreeSearch = true; // true = octree search; false = linear search
		const label cellI = probeMesh_.findCell(pI, seedCelli, useTreeSearch);
		//const label cellI = mesh().findCell(pI); // this was the old behaviour: polyMesh::findCell
                cellNos[ii] = cellI;
                
                // gather cell numbers.
                if (cellI >= 0) // if cell is found: interpolate inside it:
                {
                    UProbed[ii] = UInterp.interpolate(pI,cellI);
                    rhoProbed[ii] = rhoInterp.interpolate(pI,cellI);
                    dUdtProbed[ii] = ddtUInterp.interpolate(pI, cellI);
                }
            }
            
            
            // Gather all info from all processors. NB: based on chosen default values above (operator-dependent).
            reduce(cellNos, maxOp<List<label>>());
            reduce(UProbed, sumOp<List<vector>>());
            reduce(dUdtProbed, sumOp<List<vector>>());
            reduce(rhoProbed, maxOp<List<scalar>>());
            
            
            // do for master only:
            //Check if point was found, else fill with default values.
            if ( Pstream::master() )
            {
                int nNotFound=0;
                forAll(allPoints,ii)
                {
                    if ( cellNos[ii] < 0)
                    {
                        ++nNotFound;
                        if (probeReport_)
                            Info << "Cell not found for point: " << allPoints[ii] << endl;
                        
                        // Assign normal out-of-cfd-domain values:
                        UProbed[ii]= vector::zero;
                        rhoProbed[ii]= rhoMax; // assume cell is in water.
                        dUdtProbed[ii] = vector::zero;
                    }
                }
                if (nNotFound >0)
                    Info << nNotFound << " points were not found in any cell. The mooring probably extends outside the CFD domain." << endl;
                
                // Finally set flow in each restraint:
                motion_.allRestraints()[rI].setFluidState(motion_.time().value(),UProbed, dUdtProbed,rhoProbed);
            }
            
        } // end for all restraints
        
    }
